from .lazydeepClassifier import LazyDeepClassifier, LazyClassifier
from .lazydeepRegressor import LazyDeepRegressor, LazyRegressor
from .lazydeepMTS import LazyDeepMTS, LazyMTS

__all__ = [
    "LazyClassifier",
    "LazyRegressor",
    "LazyDeepClassifier",
    "LazyDeepRegressor",
    "LazyMTS",
    "LazyDeepMTS",
]
